/*
 * LAMIP-0.0.3 - WAV PCM input plugin
 *
 *
 */
#include <stdio.h>
#include <string.h>

#include <plug_in.h>

#define WAVE_FORMAT_UNKNOWN	(0x0000)
#define WAVE_FORMAT_PCM		(0x0001)
#define WAVE_FORMAT_ADPCM	(0x0002)
#define WAVE_FORMAT_IEEE_FLOAT	(0x0003)
#define WAVE_FORMAT_ALAW	(0x0006)
#define WAVE_FORMAT_MULAW	(0x0007)
#define WAVE_FORMAT_OKI_ADPCM	(0x0010)
#define WAVE_FORMAT_DIGISTD	(0x0015)
#define WAVE_FORMAT_DIGIFIX	(0x0016)
#define IBM_FORMAT_MULAW	(0x0101)
#define IBM_FORMAT_ALAW		(0x0102)
#define IBM_FORMAT_ADPCM	(0x0103)

#define BUFFER_LEN	(4096*6) /* for 5.1 files and 24bits files */
#define RIFF_LEN	12
#define CHUNK_LEN	8
#define FORMAT_LEN	16
#define WAVE_HEADER_LEN (RIFF_LEN + CHUNK_LEN + FORMAT_LEN + CHUNK_LEN)

#define FOURCC_EQUAL(X, a, b, c, d) (X[0]==a && X[1]==b && X[2]==c && X[3]==d)
#define WAV_MSG(s)		    lamip_send_message("WAV: " s "\n")

struct riff_t {
  unsigned char	magic[4];
  unsigned int	len;
  unsigned char	id[4];
};


struct chunk_t {
  unsigned char	id[4];
  unsigned int	len;
};

struct format_t {
  unsigned short	wFormatTag;
  unsigned short	wChannels;
  unsigned int		dwSamplesPerSec;
  unsigned int		dwAvgBytesPerSec;	/* must equal to dwSamplePerSec * wBlockAlign */
  unsigned short	wBlockAlign;		/* must equal to wChannels * (wBitsPerSample / 8) */
  unsigned short	wBitsPerSample;		/* != 0 only when wFormatTag = WAVE_FORMAT_PCM */
};

typedef struct {
  struct riff_t		riff;
  struct chunk_t	format_chunk;
  struct format_t	format;
  struct chunk_t	data_chunk;
} WaveHeader;

static int	wav_init(const lamipPluginHandle*);
static void     wav_cleanup(void);
static void     wav_decode(lamipURL*, int);

#ifdef DEBUG
static void		print_format(const struct format_t*);
#endif
static int		decoding_loop(lamipURL*, const WaveHeader*);
static unsigned short	READ_LE_U16(lamipURL*);
static unsigned int	READ_LE_U32(lamipURL*);

static InputPlugin wav_functions;


/****************************************************************************/

InputPlugin* lamip_input_info(void)
{
  wav_functions.common.name	= "wav reader";
  wav_functions.common.description = "LAMIP-0.0.3 wav PCM player";
  wav_functions.common.init	= wav_init;
  wav_functions.common.cleanup	= wav_cleanup;
  wav_functions.decode		= wav_decode;
  wav_functions.set_song_info	= 0;
  return(&wav_functions);
}

static int wav_init(const lamipPluginHandle* plug)
{
#ifdef DEBUG
  WAV_MSG("wav_init\n");
#endif
  lamip_set_mime_type((lamipPluginHandle*)plug, ".wav", 0);
  return(1);
}

static void     wav_cleanup(void)
{
#ifdef DEBUG
  WAV_MSG("wav_cleanup\n");
#endif
}

static unsigned int READ_LE_U32(lamipURL* url)
{
  unsigned char buf[4];

  lamip_url_read(url, buf, 4);
  return(buf[3] << 24 | buf[2] << 16 | buf[1] << 8 | buf[0]);
}

static unsigned short READ_LE_U16(lamipURL* url)
{
  unsigned char buf[2];

  lamip_url_read(url, buf, 2);
  return(buf[1] << 8 | buf[0]);
}

static void wav_decode(lamipURL* url, int sub_trk)
{
  //FILE*		fp;
  WaveHeader	hdr;
  lamipPCMInfo	pcmi;

#ifdef DEBUG
  WAV_MSG("wav play\n");
#endif
  lamip_url_read(url, hdr.riff.magic, sizeof(hdr.riff.magic));
  if(!FOURCC_EQUAL(hdr.riff.magic, 'R', 'I', 'F', 'F')){
    WAV_MSG("not a valid file (missing FOURCC \"RIFF\")");
    goto fail;
  }
  hdr.riff.len = READ_LE_U32(url);
  lamip_url_read(url, hdr.riff.id, sizeof(hdr.riff.id));
  if(!FOURCC_EQUAL(hdr.riff.id, 'W', 'A', 'V', 'E')){
    WAV_MSG("not a valid file (missing FOURCC \"WAVE\")");
    goto fail;
  }
  lamip_url_read(url, hdr.format_chunk.id, sizeof(hdr.format_chunk.id));
  if(!FOURCC_EQUAL(hdr.format_chunk.id, 'f', 'm', 't', ' ')){
    WAV_MSG("not a valid file (missing FOURCC \"fmt \")");
    goto fail;
  }
  hdr.format_chunk.len = READ_LE_U32(url);
  hdr.format.wFormatTag = READ_LE_U16(url);
  hdr.format.wChannels = READ_LE_U16(url);
  hdr.format.dwSamplesPerSec = READ_LE_U32(url);
  hdr.format.dwAvgBytesPerSec = READ_LE_U32(url);
  hdr.format.wBlockAlign = READ_LE_U16(url);
  hdr.format.wBitsPerSample = READ_LE_U16(url);
#ifdef DEBUG
  print_format(&hdr.format);
#endif
  if(hdr.format.wFormatTag != WAVE_FORMAT_PCM
      && hdr.format.wFormatTag != WAVE_FORMAT_IEEE_FLOAT){
    WAV_MSG("unknown format");
    goto fail;
  }
  /*** extra parameters (normaly not here for a PCM file) */
  if(hdr.format_chunk.len != FORMAT_LEN){
    char tmp[1024];
    lamip_url_read(url, tmp, hdr.format_chunk.len-16);
  }

  lamip_url_read(url, hdr.data_chunk.id, sizeof(hdr.data_chunk.id));
  if(!FOURCC_EQUAL(hdr.data_chunk.id, 'd', 'a', 't', 'a')){
    WAV_MSG("not a valid file (missing FOURCC \"data \")");
    goto fail;
  }
  hdr.data_chunk.len = READ_LE_U32(url);
  if(hdr.format.wBitsPerSample==8)
    pcmi.format = PCM_FORMAT_U8;
  else if(hdr.format.wBitsPerSample==16)
    pcmi.format = PCM_FORMAT_S16_LE;
  else if(hdr.format.wBitsPerSample==24)
    pcmi.format = PCM_FORMAT_S24_LE;
  else if(hdr.format.wBitsPerSample==32)
    pcmi.format = PCM_FORMAT_S32_LE;
  else
    pcmi.format = PCM_FORMAT_U16_LE;
  pcmi.samplerate = hdr.format.dwSamplesPerSec;
  pcmi.channels = hdr.format.wChannels;
  if(lamip_open(&pcmi, 1000*(hdr.data_chunk.len/hdr.format.dwAvgBytesPerSec))){
    if(decoding_loop(url, &hdr))
      lamip_drain();
    lamip_close();
    return;
  }
fail:
  return;
}


static int decoding_loop(lamipURL* url, const WaveHeader* hdr)
{
  unsigned char buffer[BUFFER_LEN];
  size_t	total_read = 0;
  int		ret = 0;

  /* TODO */
  if(hdr->format.wFormatTag == WAVE_FORMAT_IEEE_FLOAT){
    return(0);
  }

  while(lamip_isContinue()){
    size_t read;

    if(lamip_isSeek()){
      unsigned long ms, bytes;

      ms = lamip_isSeekGetAndReset();
      bytes = ((double) ms / 1000) * hdr->format.dwAvgBytesPerSec;
      bytes -= bytes % hdr->format.wBlockAlign;
      lamip_url_seek(url, WAVE_HEADER_LEN+bytes);
      //fseek(fp, WAVE_HEADER_LEN + bytes, SEEK_SET);
      lamip_drop();
      continue;
    }
    read = lamip_url_read(url, buffer, BUFFER_LEN);
    lamip_writeData(buffer, read);
    total_read += read;
    if(total_read >= hdr->data_chunk.len || lamip_url_eos(url)){
      ret = 1;
      break;
    }
  }
  return(ret);
}

#ifdef DEBUG
static void print_format(const struct format_t *fmt)
{
  fprintf(stderr, 
      "========================\n"
      "Format           %x\n"
      "Channels         %u\n"
      "Frequency        %u\n"
      "Bytes/sec        %u\n"
      "Block align      %u\n"
      "Bits/sample      %u\n"
      "========================\n",
      fmt->wFormatTag, fmt->wChannels, fmt->dwSamplesPerSec,
      fmt->dwAvgBytesPerSec, fmt->wBlockAlign, fmt->wBitsPerSample);
}
#endif
